//
//  AnalisadorAtaque.swift
//  Music-Mate-New-Design
//
//  Created by Raimundo Corrêa de Oliveira on 21/10/2018.
//  Copyright © 2018 LT Music Developer. All rights reserved.
//

import Foundation

/**
 * Analisador de ataque tem como proposito analisar um vetor de Float e identificar picos
 * de energia.
 *
 * @property amostrasPorUnidade define a taxa de amostragem usada para geracao do vetor de Float.
 * @property deslocarPeloAtaque define o deslocamento necessario devido ao reconhecimento de um pico.
 *
 * @constructor cria um analisador de tempo para a taxa de amostragem dada.
 */
class AnalisadorAtaque {
    //Um limite para reconhecimento de nota válida.
    private let limiar: Float = 0.45
    //Mínimo número de unidades que dois ataques podem estar separados.
    var conectividade = 2
    //Mínimo número de unidades para ataque ser considerado válido.
    var removerRuido = 3
    
    private struct Grupo {
        var comeco = 0
        var fim  = 0
        var comprimento: Int{ get { return fim - comeco + 1} }
    }
    
    /**
     * Identifica grupos de mesmos valores.
     *
     * @param entrada o vetor a ser analisado.
     * @param valorComum o valor a ser encontrado.
     */
    private func identificarGrupos(entrada: [Float], valorComum: Float) -> [Grupo] {
        
        var grupos = [Grupo]()
        var grupo = Grupo(comeco: -1, fim: -2)
        
        for index in 0..<entrada.count {
            if entrada[index] == valorComum {
                //Um valor esta proximo do outro.
                if grupo.fim == index - 1 {
                    grupo.fim = index
                }
                    //Um valor esta distante do outro.
                else {
                    //Não é o valor de inicialização
                    if (grupo.comeco != -1) { grupos.append(grupo) }
                    
                    grupo.comeco = index
                    grupo.fim = index
                }
            }
        }
        
        //Adiciona ultimo elemento se nao for o elemento de inicializacao
        if (grupo.comeco != -1) { grupos.append(grupo) }
        
        return grupos
    }
    
    /**
     * Identifica notas em um vetor de Float.
     *
     * @param entrada o vetor a ser analisado. Este deve estar normalizado.
     */
    private func identificaAtaques(entrada: [Float]) -> [(Int, Int)] {
        //Lista vazia de notas.
        var notasCruas = [(Int, Int)]()
        
        //Identifica grupos de mesmos valores.
        var ataques = identificarGrupos(entrada: entrada, valorComum: 1)
        
        //Converte os grupos identificados em notas.
        for ataque in ataques {
            notasCruas.append((ataque.comeco, ataque.fim))
        }
        
        return notasCruas
    }
    
    
    /**
     * Substitui grupos de valores comuns por outro valor.
     *
     * @param entrada o vetor a ser analisado.
     * @param valorComum o valor para identificação dos grupos.
     * @param novoValor o valor que substituirá o anterior.
     * @param tamanhoMaximo o tamanho máximo de um grupo para ser substituido.
     */
    private func substituirGrupoPorValor(entrada: inout [Float], valorComum: Float, novoValor: Float, tamanhoMaximo: Int) {
        //Os grupos de valores comuns.
        var grupos = identificarGrupos(entrada: entrada, valorComum: valorComum)
        
        //Comeca a substituicao grupo a grupo.
        for grupo in grupos {
            if (grupo.comprimento <= tamanhoMaximo) {
                //Mais um é necessário devido ao fill ser intervalo aberto no final.
                entrada.replaceSubrange(grupo.comeco...grupo.fim, with: [Float](repeating: novoValor, count: grupo.comprimento))
            }
        }
    }
    
    /**
     * Remove grupos de 1s menores ou iguais a um valor minimo.
     *
     * @param entrada o vetor a ser limpo. Deve estar binarizado.
     * @param maxTamanhoRuido a distancia entre dois 0s maxima para limpar.
     *
     * @return a entrada conectada.
     */
    private func removerRuido(entrada: inout [Float], maxTamanhoRuido: Int) {
        var tamanho = maxTamanhoRuido >= 0 ? maxTamanhoRuido : 0
        substituirGrupoPorValor(entrada: &entrada, valorComum: 1, novoValor: 0, tamanhoMaximo: tamanho)
    }
    
    /**
     * Conecta grupos de 0s menores ou iguais a um valor máximo.
     *
     * @param entrada o vetor a ser conectado. Deve estar binarizado.
     * @param distanciaMaxima a distância entre dois 1s maxima para conectar.
     *
     * @return a entrada conectada.
     */
    private func conectar(entrada: inout [Float], distanciaMaxima: Int) {
        var tamanho = distanciaMaxima >= 0 ? distanciaMaxima : 0
        substituirGrupoPorValor(entrada: &entrada, valorComum: 0, novoValor: 1, tamanhoMaximo: tamanho)
    }
    
    /**
     * Binariza em 0s e 1s um vetor de Float com base em um valor.
     *
     * @param entrada o vetor a ser binarizado.
     * @param valor o valor base para binarizacao.
     *
     * @return o vetor binarizado.
     */
    private func binarizar(entrada: inout [Float], valor: Float) {
        //Valores menores que um valor viram 0, demais valores viram 1.
        for index in 0..<entrada.count {
            if entrada[index] >= valor {
                entrada[index] = 1
            } else {
                entrada[index] = 0
            }
        }
    }
    
    /**
     * Identifica ataques em um vetor de Float.
     *
     * @param entrada o vetor a ser analisado. Este deve estar normalizado.
     */
    private func extrairAtaques(entrada: [Float]) -> [(Int, Int)] {
        var aux = entrada
        //Binariza a entrada com 0s e 1s.
        binarizar(entrada: &aux, valor: limiar)
        //Conecta a entrada binarizada com a distância mínima dada.
        conectar(entrada: &aux, distanciaMaxima: conectividade - 1)
        //Remove notas menores que o valor dado.
        removerRuido(entrada: &aux, maxTamanhoRuido: removerRuido - 1)
        
        return identificaAtaques(entrada: aux)
    }
    
    /**
     * Identifica ataques em um vetor de Float.
     *
     * @param entrada o vetor a ser analisado.
     *
     * @return a lista de ataques identificados como par de inteiros.
     * O vetor de Pair contém 2 valores: unidade inicial e unidade final, nessa ordem.
     */
    func ataquesEm(entrada: [Float]) -> [(Int, Int)] {
        //Se entrada for vazia não há ataques.
        if (entrada.isEmpty) { return [(Int, Int)]() }
        
        //Identifica os ataques na entrada.
        return extrairAtaques(entrada: Vetor().normalizar(entrada: entrada))
    }
}
