//
//  AnalisadorFrequencia.swift
//  Music-Mate-New-Design
//
//  Created by LTMD on 23/08/2018.
//  Copyright © 2018 LT Music Developer. All rights reserved.
//

import Foundation

/**
 * Analisador de frequencia tem como proposito analisar um vetor de Float e
 * identificar as frequencias que aparecem dentro dele.
 *
 * @property taxaDeAmostragem define a taxa de amostragem usado nos cálculos.
 * @property comprimentoFourier define a potência de 2 que sera transformada.
 * @property freqRange define as configuracao necessarias para conversao de frequencia em nota.
 * @property valorEnergiaDeDescarte define a energia mínima que um pico deve ter.
 * @property dbLimitante define a diferença mínima entre a maior energia e um energia qualquer.
 *
 * @constructor cria um analisador de frequencia para as configuracoes dadas.
 */
class AnalisadorFrequencia {
    private let taxaDeAmostragem: Int
    private let comprimentoFourier: Int
    
    //Os parametros de taxaAmostragem e comprimento
    //influenciam muito na conversão de frequencia para nota.
    private let freqRange: FreqRange
    
    //A classe responsável pelo processamento de sinal.
    private let dsp: DSP
    
    //A classe responsável pelo processamento de sinal.
    private let valorEnergiaDeDescarte: Float = 10
    
    //Limite para descarte na comparação dos picos.
    private let dbLimitante: Float = -8
    
    init(taxaDeAmostragem: Int, comprimentoFourier: Int) {
        self.taxaDeAmostragem = taxaDeAmostragem
        self.comprimentoFourier = comprimentoFourier
        
        freqRange = FreqRange(taxaAmostragem: taxaDeAmostragem, comprimentoFourier: comprimentoFourier)
        
        dsp = DSP(comprimentoFourier: comprimentoFourier)
    }
    
    /**
     * Remove picos com mesmo filtro preservando maior energia.
     *
     * @param picos os picos a serem tratados.
     */
    private func maximizarAmplitude(picos: inout [Pico]) {
        //Ordena de forma crescente por filtro.
        picos.sort(by:) { $0.filtro < $1.filtro }
        
        //A lista onde será guardado os maiores picos.
        var novaLista = [Pico]()
        
        //O pico encontrado com maior energia.
        var maiorPico = Pico()
        
        picos.forEach {
            //Se for a mesmo filtro guarda maior energia,
            //senão adiciona a lista e analisa outro pico.
            if maiorPico.filtro != $0.filtro {
                if !maiorPico.estaVazio {
                    novaLista.append(maiorPico.copy())
                }
                
                maiorPico = $0.copy()
            } else {
                maiorPico.energia = max(maiorPico.energia, $0.energia)
            }
        }
        
        //Adiciona o ultimo pico a lista.
        if (!maiorPico.estaVazio) {
            novaLista.append(maiorPico)
        }
        
        //Limpa a lista de entra.
        picos.removeAll()
        
        //Atualiza a lista de entrada.
        picos.append(contentsOf: novaLista)
    }
    
    /**
     * Encontra picos de valores em um vetor.
     *
     * @param entrada o vetor a ser analisado.
     *hr
     * @return os picos encontrados no array.
     */
    private func localizarPicos(entrada: [Float]) -> [Pico] {
        //A lista onde será guardado os picos.
        var picos = [Pico]()
        
        for i in 1..<(entrada.count / 2) {
            //Verifica se o valor anterior e o próximo sõo menores que o atual.
            if (entrada[i-1] < entrada[i] && entrada[i] > entrada[i+1]) {
                //Se o valor encontrado for menor que 10 pode ser descartado.
                if (entrada[i] * entrada[i] > valorEnergiaDeDescarte) {
                    picos.append(Pico(index: i, filtro: freqRange.notaDe(amostra: i), energia: entrada[i] * entrada[i]))
                }
            }
        }
        
        //Remove picos com mesmo filtro.
        maximizarAmplitude(picos: &picos)
        
        return picos
    }
    
    /**
     * Remove picos com mesmo filtro somando as energias.
     *
     * @param picos os picos sem repetição de filtro.
     */
    func somaAmplitude(picos: inout [Pico]) {
        //Ordena de forma crescente por filtro.
        picos.sort(by:) { $0.filtro < $1.filtro }
        
        //A lista onde será guardado a soma dos picos.
        var novaLista = [Pico]()
        
        //O pico resultante da soma.
        var picoAux = Pico()
        
        picos.forEach {
            //Se for a mesmo filtro adiciona a energia,
            //senão adiciona a lista e analisa outro pico.
            if (picoAux.filtro != $0.filtro) {
                if (!picoAux.estaVazio) {
                    novaLista.append(picoAux.copy())
                }
                picoAux = $0.copy()
            } else {
                picoAux.energia += $0.energia
            }
        }
        
        //Adiciona a ultima nota a lista.
        if (!picoAux.estaVazio) {
            novaLista.append(picoAux)
        }
        
        //Limpa a lista de entra.
        picos.removeAll()
        
        //Atualiza a lista de entrada.
        picos.append(contentsOf: novaLista)
    }
    
    /**
     * Identifica picos de energia em um sinal de entrada.
     *
     * @param entrada indica os dados a serem processados.
     */
    private func picosDeEnergiaEm(entrada: [Float]) -> [Pico] {
        let janelas = Vetor().dividir(entrada: entrada, compPedaco: comprimentoFourier, passo: nil)
        
        var picos = [Pico]()
        
        janelas.forEach { picos.append(contentsOf: localizarPicos(entrada: dsp.transformar(signal: $0))) }
        
        somaAmplitude(picos: &picos)
        
        return picos
    }
    
    /**
     * Elimina picos com energia em db menor que um valor, em relacao ao maior.
     *
     * @param picos os picos a serem analisados.
     * @param limiteDb o limite de db para eliminacao de pico.
     */
    func dbDiference(picos: inout [Pico], limiteDb: Float) {
        //Ordena em ordem decrescente de energia.
        picos.sort(by:) { $0.energia > $1.energia }
        
        //A lista onde sera guardado os picos validos.
        var novaLista = [Pico]()
        
        //O maior pico.
        let picoMax = picos.first ?? Pico()
        
        picos.forEach {
            //Calcula a diferenca de energia em db.
            let db = 10 * log10f($0.energia / picoMax.energia)
            
            if db >= limiteDb {
                novaLista.append($0.copy())
            }
        }
        
        //Limpa a lista de entra.
        picos.removeAll()
        
        //Atualiza a lista de entrada.
        picos.append(contentsOf: novaLista)
    }
    
    /**
     * Identifica as notas que foram tocas no vetor de entrada.
     *
     * @param entrada indica o vetor a ser analisado.
     */
    func frequenciasEm(entrada: [Float]) -> [Pico] {
        var picosDeEnergia = picosDeEnergiaEm(entrada: entrada)
        dbDiference(picos: &picosDeEnergia, limiteDb: dbLimitante)
        
        return picosDeEnergia
    }
}
