//
//  Vector.swift
//  Engine
//
//  Created by Diego Lopes on 4/4/18.
//  Copyright © 2018 Diego Developer. All rights reserved.
//

import UIKit

class Vetor {
    /**
     * Divide um vetor em pedacos com o mesmo comprimento.
     *
     * @param entrada o vetor a ser dividido.
     * @param compPedaco o comprimento de cada pedaco.
     * @param passo adistancia entre os comecos de dois pedacos consecutivos.
     *
     * @return os pedacos do vetor.
     */
    func dividir(entrada: [Float], compPedaco: Int, passo: Int?) -> [[Float]] {
        var pedacos = [[Float]]()
        
        if (compPedaco <= 0 || entrada.count <= 0) {
            return pedacos
        }
        
        //Se passo for nulo, tera o mesmo valor de compPedaco
        var passoAux = passo ?? 0
        
        if (passoAux <= 0) {
            passoAux = compPedaco
        }
        
        var comeco = -passoAux
        var fim = compPedaco - passoAux - 1
        
        repeat {
            comeco += passoAux
            fim += passoAux
            var vetor = [Float](repeating: 0, count: fim - comeco + 1)
            for i in 0..<vetor.count {
                if comeco + i < entrada.count {
                    vetor[i] = entrada[comeco + i]
                }
            }
            pedacos.append(vetor)
        } while (fim < entrada.count - 1)
        
        return pedacos
    }
    
    /**
     * Normaliza um vetor.
     *
     * @param entrada o vetor a ser normalizado.
     *
     * @return o vetor normalizado.
     */
    func normalizar(entrada: [Float]) -> [Float] {
        let highestValue = maiorValorAbsoluto(vetor: entrada, range: nil)
        return escalar(entrada: entrada, valor: Float(1) / highestValue)
    }
    
    /**
     * Escala um vetor.
     *
     * @param entrada o vetor a ser escalado.
     * @param valor a escala a usar.
     *
     * @return o vetor escalado.
     */
    func escalar(entrada: [Float], valor: Float) -> [Float] {
        return entrada.map{ $0 * valor }
    }
    
    /**
     * Cria um vetor complexo para uma entrada real.
     *
     * @param entrada o vetor real.
     *
     * @return o vetor complexo.
     */
    func complexo(entrada: [Float]) -> [Float] {
        var saida = [Float](repeating: 0, count: entrada.count * 2)
        for index in 0..<entrada.count {
            saida[index * 2] = entrada[index]
        }
        return saida
    }
    
    /**
     * Calcula a magnitude de um vetor complexo.
     *
     * @param entrada o vetor complexo.
     *
     * @return o vetor real.
     */
    func magnitude(entrada: [Float]) -> [Float] {
        var saida = [Float](repeating: 0, count: entrada.count / 2)
        for index in 0..<saida.count {
            saida[index] = Complexo().magnitude(real: entrada[index * 2], imag: entrada[index * 2 + 1])
        }
        return saida
    }
    
    /**
     * Calculates the absolute values of a vector
     *
     * @param source The source vector.
     * @param output The vector to store the absolute values.
     */
    func absolute(source: [Float]) -> [Float] {
        return source.map { abs($0) }
    }
    
    /**
     * Calcula o maior valor absoluto em um range.
     *
     * @param vetor o vetor a ser analisado.
     * @param range o range a ser analisado.
     *
     * @return o maior valor no range.
     */
    func maiorValorAbsoluto(vetor: [Float], range: [Int]?) -> Float {
        let valid = validRange(vector: vetor, range: range)
        
        var aux = [Float](repeating: 0, count: valid[1] - valid[0] + 1)
        for i in valid[0]...valid[1] {
            aux[i - valid[0]] = vetor[i]
        }
        return absolute(source: aux).max() ?? 0
    }
    
    
    /**
     * Calculates a valid range based in a vector and a given range.
     *
     * @param vector    The vector to be analyzed.
     * @param range     The range to be analyzed.
     * @return A valid range or the range it self if already valid.
     */
    private func validRange(vector: [Float], range: [Int]?) -> [Int] {
        //If null range becomes the vector range.
        if let rangeAux = range {
            if (rangeAux[0] > rangeAux[1]) {
                return [rangeAux[1], rangeAux[0]]
            } else if (rangeAux[0] < 0 || rangeAux[0] >= vector.count) {
                return [0, 0]
            } else if (rangeAux[1] >= vector.count) {
                return [rangeAux[0], vector.count - 1]
            } else {
                return rangeAux
            }
        }else {
            return [0, vector.count - 1]
        }
    }
    
    /**
     * Desloca o vetor para a direita.
     *
     * @param entrada o vetor a ser deslocado.
     * @param quantidade a quantidade a deslocar.
     *
     * @return o vetor deslocado.
     */
    func shtDireita(entrada: [Float], quantidade: Int) -> [Float] {
        var saida = [Float](repeating: 0, count: entrada.count)
        for i in quantidade..<entrada.count {
            saida[quantidade] = entrada[i - quantidade]
        }
        return saida
    }
    
    /**
     * Desloca o vetor para a esquerda.
     *
     * @param entrada o vetor a ser deslocado.
     * @param quantidade a quantidade a deslocar.
     *
     * @return o vetor deslocado.
     */
    func shtEsquerda(entrada: [Float], quantidade: Int) -> [Float] {
        var saida = [Float](repeating: 0, count: entrada.count + quantidade - quantidade)
        for i in quantidade..<entrada.count {
            saida[i - quantidade] = entrada[i]
        }
        return saida
    }
}
