//
//  FreqRange.swift
//  Music-Mate-New-Design
//
//  Created by LTMD on 23/08/2018.
//  Copyright © 2018 LT Music Developer. All rights reserved.
//

import Foundation

/**
 * Gera intervalos de frequencias valor amostras para cada umas das notas
 * de C1 a B7.
 *
 * @param taxaAmostragem a taxa de amostragem. Se nula,
 *                          nao instancia os intervalos de radianos nem de amostras.
 * @param compFourier o comprimento de fourier. Se nulo,
 *                          nao instancia os intervalos de amostras.
 */
class FreqRange {
    private let taxaDeAmostragem: Int
    private let comprimentoFourier: Int
    
    private let numNotas = 84
    
    //Raiz decima segunda de 2.
    private let r: Float = pow(2.0, 1.0/12.0)
    
    //Razao entre 2 * pi valor a taxa de amostragem.
    private lazy var pi2s = (taxaDeAmostragem != 0) ? (2.0 * Float.pi / Float(taxaDeAmostragem)) : 0.0
    //Razao entre compriemnto de fourier valor taxa de amostragem.
    private lazy var NFs = (taxaDeAmostragem != 0) ? (Float(comprimentoFourier) / Float(taxaDeAmostragem)) : 0.0
    
    //Menor frequencia inferior.
    private let frequenciaInicial: Float = 31.7308794
    //    private let frequenciaInicial: Float = 30.676
//        private let frequenciaInicial: Float = 30.4
    //O intervalo de frequencia em Hz para cada nota.
    private lazy var intervaloHz = [(Float, Float)](repeating: (0.0, 0.0), count: numNotas)
    //O intervalo de frequencia em radianos para cada nota.
    private lazy var intervaloRad = [(Float, Float)](repeating: (0.0, 0.0), count: numNotas)
    
    //Calcula os intervalos.
    init(taxaAmostragem: Int, comprimentoFourier: Int) {
        self.taxaDeAmostragem = taxaAmostragem
        self.comprimentoFourier = comprimentoFourier
        
        calculateRanges()
    }
    
    /**
     * Calculas os intervalos de frequencia para cada nota.
     */
    private func calculateRanges() {
        for nota in 0..<numNotas {
            intervaloHz[nota] = freqIntervalo(nota)
            intervaloRad[nota] = omegaIntervalo(intervaloHz[nota])
        }
    }
    
    /**
     * Gera um intervalo de frequencias em Hz para uma nota.
     *
     * @param nota a nota a ser usada.
     *
     * @return um intervalo de frequencia para a nota dada.
     */
    private func freqIntervalo(_ nota: Int) -> (Float, Float) {
        //Calculo de intervalo.
        let intInferior = frequenciaInicial * pow(r, Float(nota))
        let intSuperior = frequenciaInicial * pow(r, Float(nota + 1))
        
        return (intInferior, intSuperior)
    }
    
    /**
     * Calcula o intervalo de frequencia angular normalizada para o intervalo de frequencia dado.
     *
     * @param freqIntervalo o intervalo de frequencia em Hz.
     *
     * @return o intervalo em frequencia angular normalizado.
     */
    private func omegaIntervalo(_ freqIntervalo: (Float, Float)) -> (Float, Float) {
        //Calculo de intervalo.
        let omegaI = pi2s * freqIntervalo.0
        let omegaF = pi2s * freqIntervalo.1
        
        return (omegaI, omegaF)
    }
    
    /**
     * Gera um intervalo de amostras para um intervalo de frequencia.
     *
     * @param freqIntervalo o intervalo em Hz.
     *
     * @return um intervalo de amostras para as frequencias dadas.
     */
    private func amostIntervalo(_ freqIntervalo: (Float, Float)) -> (Int, Int) {
        //Calculo de intervalo.
        let amostraI = NFs * freqIntervalo.0
        let amostraF = NFs * freqIntervalo.1
        
        return (Int(amostraI.rounded()), Int(amostraF.rounded()))
    }
    
    /**
     * Retorna o intervalo de frequencia em Hz para nota.
     *
     * @param nota a nota a ser analisada.
     *
     * @return o intervalo para a nota data.
     */
    func intervaloHz(nota: Int) -> (Float, Float) {
        return (nota >= 0 && nota < numNotas) ? intervaloHz[nota] : (0.0, 0.0)
    }
    
    /**
     * Retorna o intervalo de frequencia em radianos para nota.
     *
     * @param nota a nota a ser analisada.
     *
     * @return o intervalo para a nota data.
     */
    func intervaloRad(nota: Int) -> (Float, Float) {
        return (nota >= 0 && nota < numNotas) ? intervaloRad[nota] : (0.0, 0.0)
    }
    
    /**
     * Retorna a nota que ocorre na amostra dada.
     *
     * @param amostra a amostra a ser analisada.
     */
    func notaDe(amostra: Int) -> Int {
        let cima = log(Float(amostra) / (NFs * frequenciaInicial))
        let baixo = log(r)
        let nota = Int(floor(cima / baixo))
        
        return (nota >= 0 && nota < numNotas) ? nota : 0
    }
}
